/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <string.h>
#include <ctype.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <errno.h>
#include <dlfcn.h>
#include <dirent.h>

#ifdef RTCTIMER_LINUX
#include <linux/rtc.h>
#include <sys/poll.h>
#include <sys/ioctl.h>
#endif

#if defined(__FreeBSD__)
#include <machine/param.h>
#endif

#include "../qcommon/qcommon.h"
#include "glob.h"

cvar_t *nostdout;

unsigned	sys_frame_time;

uid_t saved_euid;
qboolean stdin_active = qtrue;

#ifdef RTCTIMER_LINUX
/* needed for RTC timer */
static int rtc_rate = -1;
static int rtc_fd = 0;  /* file descriptor for rtc device */
#endif

// =======================================================================
// General routines
// =======================================================================

void Sys_ConsoleOutput (char *string)
{
	if (nostdout && nostdout->integer)
		return;

	fputs (string, stdout);
}

/*
=================
Sys_Quit
=================
*/
void Sys_Quit (void)
{
	fcntl (0, F_SETFL, fcntl (0, F_GETFL, 0) & ~FNDELAY);

	Qcommon_Shutdown ();
#ifdef RTCTIMER_LINUX
	if (rtc_fd) {
		close(rtc_fd);
		rtc_fd = 0;
	}
#endif

	_exit(0);
}

/*
=================
Sys_Init
=================
*/
void Sys_Init(void)
{
}

/*
================
Sys_InitDynvars
================
*/
void Sys_InitDynvars (void)
{
}

/*
=================
Sys_Error
=================
*/
void Sys_Error (char *error, ...)
{ 
	va_list     argptr;
	char        string[1024];

// change stdin to non blocking
	fcntl (0, F_SETFL, fcntl (0, F_GETFL, 0) & ~FNDELAY);

	CL_Shutdown ();
	
	va_start (argptr, error);
	vsnprintf (string, 1024, error, argptr);
	va_end (argptr);
	fprintf (stderr, "Error: %s\n", string);

	Qcommon_Shutdown ();
#ifdef RTCTIMER_LINUX
	if (rtc_fd) {
		close(rtc_fd);
		rtc_fd = 0;
	}
#endif

	_exit (1);
} 

/*
================
Sys_Milliseconds
================
*/

#ifdef RTCTIMER_LINUX
/* RTC timer code is adapted from EzQuake */
static quint64 Microseconds_RTC (void)
{
	/* rtc timer vars */
	unsigned long curticks = 0;
	struct pollfd pfd;
	static unsigned long totalticks;

	pfd.fd = rtc_fd;
	pfd.events = POLLIN | POLLERR;

	while (poll(&pfd, 1, 100000) < 0) {
		if ((errno = EINTR))
			continue; /* happens with gdb or signal exiting */
		else
			Sys_Error("Poll call on RTC timer failed!\n");
	}

	read(rtc_fd, &curticks, sizeof(curticks));
	curticks = curticks >> 8; /* knock out info byte */
	totalticks += curticks;
	
	// TODO handle the wrap
	return totalticks / (rtc_rate/1000000.0);
}
#endif

static quint64 Microseconds_GTOD (void)
{
	struct timeval tp;
	struct timezone tzp;
	static unsigned int		secbase;

	gettimeofday(&tp, &tzp);
	
	if (!secbase)
	{
		secbase = tp.tv_sec;
		return tp.tv_usec;
	}

	// TODO handle the wrap
	return (quint64)(tp.tv_sec - secbase)*1000000 + tp.tv_usec;
}

quint64 Sys_Microseconds (void)
{
#ifdef RTCTIMER_LINUX
	if (rtc_fd)
		return Microseconds_RTC();
	else
#endif
		return Microseconds_GTOD();
}

unsigned int Sys_Milliseconds (void)
{
	return Sys_Microseconds() / 1000;
}

qboolean Sys_CreateDirectory( const char *path )
{
	return (!mkdir(path, 0777));
}

qboolean Sys_RemoveDirectory( const char *path )
{
	return (!rmdir(path));
}

void Sys_Sleep (unsigned int millis)
{
	usleep(millis * 1000);
}

void floating_point_exception_handler (int whatever)
{
	signal (SIGFPE, floating_point_exception_handler);
}

char *Sys_ConsoleInput(void)
{
	static char text[256];
	int     len;
	fd_set	fdset;
	struct timeval timeout;

	if (!dedicated || !dedicated->integer)
		return NULL;

	if (!stdin_active)
		return NULL;

	FD_ZERO (&fdset);
	FD_SET (0, &fdset); // stdin
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
	if (select (1, &fdset, NULL, NULL, &timeout) == -1 || !FD_ISSET(0, &fdset))
		return NULL;

	len = read (0, text, sizeof(text));
	if (len == 0) { // eof!
		stdin_active = qfalse;
		return NULL;
	}

	if (len < 1)
		return NULL;

	text[len-1] = 0;    // rip off the /n and terminate

	return text;
}

/*
=================
Sys_GetSymbol
=================
*/
#ifdef SYS_SYMBOL
void *Sys_GetSymbol( const char *moduleName, const char *symbolName ) {
	// FIXME: Does not work on Debian64 for unknown reasons (dlsym always returns NULL)
	void * const module = dlopen(moduleName, RTLD_NOW);
	if (module) {
		void * const symbol = dlsym(module, symbolName);
		dlclose(module);
		return symbol;
	} else
		return NULL;
}
#endif // SYS_SYMBOL

/*
========================================================================

DLL

========================================================================
*/

/*
=================
Sys_UnloadLibrary
=================
*/
void Sys_UnloadLibrary( void **lib )
{
	if( lib && *lib ) {
		if( dlclose( *lib ) )
			Com_Error( ERR_FATAL, "dlclose failed" );
		*lib = NULL;
	}
}

/*
=================
Sys_LoadLibrary
=================
*/
void *Sys_LoadLibrary( char *name, dllfunc_t *funcs )
{
	void *lib;
	dllfunc_t *func;

	if( !name || !name[0] || !funcs )
		return NULL;

	Com_DPrintf( "LoadLibrary (%s)\n", name );

	lib = dlopen( name, RTLD_NOW );
	if( !lib )
	{
		// wsw : pb : error handling
		Com_Printf( "LoadLibrary (%s):(%s)\n", name, dlerror() );
		return NULL;
	}
	for( func = funcs; func->name; func++ ) {
		*(func->funcPointer) = ( void * )dlsym( lib, func->name );

		if( !(*(func->funcPointer)) ) {
			Sys_UnloadLibrary( &lib );
			Com_Error( ERR_FATAL, "%s: dlsym failed for %s", name, func->name );
		}
	}

	return lib;
}

/*****************************************************************************/

static void *game_library = NULL;
static char *game_fullname = NULL;
static void *cgame_library = NULL;
static char *cgame_fullname = NULL;
static void *ui_library = NULL;
static char *ui_fullname = NULL;

static void *sound_library = NULL;

#ifdef __cplusplus
# define EXTERN_API_FUNC	extern "C"
#else
# define EXTERN_API_FUNC	extern
#endif

/*
=================
Sys_UnloadGameLibrary
=================
*/
void Sys_UnloadGameLibrary( gamelib_t gamelib )
{
	// Prevent compiler warning
	void **lib = NULL;
	char *libname = NULL;
	char *fullname = NULL;
	char *p;

	switch( gamelib ) {
		case LIB_GAME:
			lib = &game_library;
#ifdef GAME_HARD_LINKED
			*lib = NULL;
#endif
			libname = "game_" ARCH ".so";
			fullname = game_fullname;
			game_fullname = NULL;
			break;
		case LIB_CGAME:
			lib = &cgame_library;
#ifdef CGAME_HARD_LINKED
			*lib = NULL;
#endif
			libname = "cgame_" ARCH ".so";
			fullname = cgame_fullname;
			cgame_fullname = NULL;
			break;
		case LIB_UI:
			lib = &ui_library;
#ifdef UI_HARD_LINKED
			*lib = NULL;
#endif
			libname = "ui_" ARCH ".so";
			fullname = ui_fullname;
			ui_fullname = NULL;
			break;
		default:
			assert( qfalse );
	}

	if( *lib ) {
		if( dlclose(*lib) )
			Com_Error( ERR_FATAL, "dlclose failed" );
		*lib = NULL;

		FS_RemoveAbsoluteFile( fullname );
		p = strrchr( fullname, '/' );
		if( p ) {
			*p = '\0';
			FS_RemoveAbsoluteDirectory( fullname );
		}
		Mem_ZoneFree( fullname );
	}
}

/*
=================
Sys_LoadGameLibrary
=================
*/
void *Sys_LoadGameLibrary( gamelib_t gamelib, qboolean pure, void *parms )
{
	static int randomizer = 0; // random part of tempmodules dir, always the same for one launch of Warsow
	char *tempname;
	int tempname_size;
	void *(*APIfunc)(void *);
	void **lib = NULL;
	char *libname = NULL;
	char *apifuncname = NULL;

	if( !randomizer ) {
		srand( time(0) );
		randomizer = brandom( 0, 9999 );
	}

	APIfunc = NULL;
	switch( gamelib ) {
		case LIB_GAME:
		{
#ifdef GAME_HARD_LINKED
			EXTERN_API_FUNC void *GetGameAPI( void * );
			APIfunc = GetGameAPI;
#endif
			lib = &game_library;
			libname = "game_" ARCH ".so";
			apifuncname = "GetGameAPI";
			break;
		}
		case LIB_CGAME:
		{
#ifdef CGAME_HARD_LINKED
			EXTERN_API_FUNC void *GetCGameAPI( void * );
			APIfunc = GetCGameAPI;
#endif
			lib = &cgame_library;
			libname = "cgame_" ARCH ".so";
			apifuncname = "GetCGameAPI";
			break;
		}
		case LIB_UI:
		{
#ifdef UI_HARD_LINKED
			EXTERN_API_FUNC void *GetUIAPI( void * );
			APIfunc = GetUIAPI;
#endif
			lib = &ui_library;
			libname = "ui_" ARCH ".so";
			apifuncname = "GetUIAPI";
			break;
		}
		default:
			assert( qfalse );
	}

	if( *lib )
		Com_Error( ERR_FATAL, "Sys_LoadGameLibrary without Sys_UnloadGameLibrary" );

	if( APIfunc ) {
		*lib = ( void * )1;
		return APIfunc( parms );
	}

	// it exists?
	if( FS_FOpenFile(libname, NULL, FS_READ) == -1 ) {
		Com_Printf( "LoadLibrary (%s):(File not found)\n", libname );
		return NULL;
	}

	// pure check
	if( pure ) {
		if( !FS_IsPureFile(libname) ) {
			Com_Printf( "LoadLibrary (%s):(Unpure file)\n", libname );
			return NULL;
		}
	}

	tempname_size = strlen(FS_WriteDirectory()) + 1 + strlen(FS_GameDirectory()) + strlen("/tempmodules") +
		strlen(va("%i", randomizer)) + 1 + strlen(libname) + 1;
	tempname = Mem_ZoneMalloc( tempname_size );

	Q_snprintfz( tempname, tempname_size, "tempmodules%i/%s", randomizer, libname ); // without gamedir for copy

	if( !FS_CopyFile(libname, tempname) ) {
		Com_Printf( "LoadLibrary (%s):(FS_CopyFile failed)\n", libname );
		Mem_ZoneFree( tempname );
		return NULL;
	}

	// with gamedir for dlopen
	Q_snprintfz( tempname, tempname_size, "%s/%s/tempmodules%i/%s", FS_WriteDirectory(), FS_GameDirectory(),
		randomizer, libname );

	*lib = dlopen( tempname, RTLD_NOW );

	if( !(*lib) ) {
		Com_Printf( "LoadLibrary (%s):(%s)\n", tempname, dlerror() );
		Mem_ZoneFree( tempname );
		return NULL;
	}

	APIfunc = (void *)dlsym( *lib, apifuncname );
	if( !APIfunc ) {
		Com_Printf( "LoadLibrary (%s):(%s)\n", tempname, dlerror() );
		Mem_ZoneFree( tempname );
		Sys_UnloadGameLibrary( gamelib );
		return NULL;
	}

	switch( gamelib ) {
		case LIB_GAME:
			game_fullname = tempname;
			break;
		case LIB_CGAME:
			cgame_fullname = tempname;
			break;
		case LIB_UI:
			ui_fullname = tempname;
			break;
		default:
			assert( qfalse );
	}

	return APIfunc( parms );
}

/*
=================
Sys_UnloadSoundLibrary
=================
*/
void Sys_UnloadSoundLibrary( void )
{
	// Prevent compiler warning
	void **lib;

	lib = &sound_library;
#ifdef SOUND_HARD_LINKED
	*lib = NULL;
#endif

	if( *lib ) {
		if( dlclose(*lib) )
			Com_Error( ERR_FATAL, "dlclose failed" );
		*lib = NULL;
	}
}
/*
=================
Sys_LoadSoundLibrary
=================
*/
void *Sys_LoadSoundLibrary( const char *basename, void *parms )
{
	void *(*APIfunc)(void *);
	void **lib = NULL;
	char *apifuncname = NULL;
	char libname[MAX_QPATH];
	const char *fullname;

	APIfunc = NULL;

#ifdef SOUND_HARD_LINKED
	EXTERN_API_FUNC void *GetGameAPI( void * );
	APIfunc = GetSoundAPI;
#endif
	lib = &sound_library;
	Q_snprintfz( libname, sizeof(libname), "libs/snd_%s_%s.so", basename, ARCH );
	apifuncname = "GetSoundAPI";

	if( *lib )
		Com_Error( ERR_FATAL, "Sys_LoadSoundLibrary without Sys_UnloadSoundLibrary" );

	if( APIfunc ) {
		*lib = ( void * )1;
		return APIfunc( parms );
	}

	fullname = FS_AbsoluteNameForBaseFile( libname );

	if( !fullname ) {
		Com_Printf( "LoadLibrary (%s):(Not found)\n", libname );
		return NULL;
	}

	*lib = dlopen( fullname, RTLD_NOW );

	if( !(*lib) ) {
		Com_Printf( "LoadLibrary (%s):(%s)\n", fullname, dlerror() );
		return NULL;
	}

	APIfunc = (void *)dlsym( *lib, apifuncname );
	if( !APIfunc ) {
		Com_Printf( "LoadLibrary (%s):(%s)\n", fullname, dlerror() );
		Sys_UnloadSoundLibrary();
		return NULL;
	}

	return APIfunc( parms );
}

//===============================================================================

static char			*findbase = NULL;
static const char	*findpattern = NULL;
static char			*findpath = NULL;
static int			findpath_size = 0;
static DIR			*fdir = NULL;

static qboolean CompareAttributes( const char *name, unsigned musthave, unsigned canthave )
{
	struct stat st;

	assert( name );

	// . and .. never match
	if( !strcmp(COM_FileBase(name), ".") || !strcmp(COM_FileBase(name), "..") )
		return qfalse;

	if( stat(name, &st) == -1 ) {
		Com_Printf( "Warning, stat failed: %s\n", name );
		return qfalse; // shouldn't happen
	}

	if( (st.st_mode & S_IFDIR) && (canthave & SFF_SUBDIR) )
		return qfalse;

	if( (musthave & SFF_SUBDIR) && !(st.st_mode & S_IFDIR) )
		return qfalse;

	return qtrue;
}

const char *Sys_FindFirst( const char *path, unsigned musthave, unsigned canhave )
{
	char *p;

	assert( path );
	assert( !fdir );
	assert( !findbase && !findpattern && !findpath && !findpath_size );

	if( fdir )
		Sys_Error( "Sys_BeginFind without close" );

	findbase = Mem_TempMalloc( sizeof(char) * (strlen(path) + 1) );
	Q_strncpyz( findbase, path, sizeof(char) * (strlen(path) + 1) );

	if( (p = strrchr(findbase, '/')) ) {
		*p = 0;
		if( !strcmp(p+1, "*.*") ) // *.* to *
			*(p+2) = 0;
		findpattern = p+1;
	} else {
		findpattern = "*";
	}
	
	if( !(fdir = opendir(findbase)) )
		return NULL;

	return Sys_FindNext( musthave, canhave );
}

const char *Sys_FindNext( unsigned musthave, unsigned canhave )
{
	struct dirent *d;
	int size;

	assert( fdir );
	assert( findbase && findpattern );

	if( !fdir )
		return NULL;

	while( (d = readdir(fdir)) )
	{
		if( !*findpattern || glob_match(findpattern, d->d_name) )
		{
			size = sizeof(char) * (strlen(findbase) + 1 + strlen(d->d_name) + 1);
			if( findpath_size < size ) {
				if( findpath )
					Mem_TempFree( findpath );
				findpath_size = size * 2; // extra size to reduce reallocs
				findpath = Mem_TempMalloc( findpath_size );
			}
			Q_snprintfz( findpath, findpath_size, "%s/%s", findbase, d->d_name );

			if( CompareAttributes(findpath, musthave, canhave) )
				return findpath;
		}
	}
	return NULL;
}

void Sys_FindClose( void )
{
	assert( findbase );

	if( fdir ) {
		closedir( fdir );
		fdir = NULL;
	}

	Mem_TempFree( findbase );
	findbase = NULL;
	findpattern = NULL;

	if( findpath ) {
		Mem_TempFree( findpath );
		findpath = NULL;
		findpath_size = 0;
	}
}

/*
=================
Sys_GetHomeDirectory
=================
*/
const char *Sys_GetHomeDirectory( void )
{
	return getenv( "HOME" );
}

//===============================================================================

/*
=================
Sys_GetClipboardData
=================
*/

char *Sys_GetClipboardData(void)
{
#if 0
	Window sowner;
	Atom type, property;
	unsigned long len, bytes_left, tmp;
	unsigned char *data;
	int format, result;
	char *ret = NULL;
		
	if (!x11display.dpy && x11display.win) 
		return NULL;
	
	sowner = XGetSelectionOwner (x11display.dpy, XA_PRIMARY);
			
	if (sowner != None)
	{
		property = XInternAtom(x11display.dpy, "GETCLIPBOARDDATA_PROP", False);
				
		XConvertSelection (x11display.dpy, XA_PRIMARY, XA_STRING, property, x11display.win, myxtime); /* myxtime == time of last X event */
		XFlush (x11display.dpy);

		XGetWindowProperty (x11display.dpy, x11display.win, property,  0, 0, False, AnyPropertyType, &type, &format, &len, &bytes_left, &data);

		if (bytes_left > 0) 
		{
			result = XGetWindowProperty(x11display.dpy, x11display.win, property,
				0, bytes_left, True, AnyPropertyType,  &type, &format, &len, &tmp, &data);

			if (result == Success){
				ret = strdup((char *) data);
			}

			XFree(data);
		}
	}

	return ret;
#endif
	return NULL;
}

/*
=================
Sys_AppActivate
=================
*/
void Sys_AppActivate (void)
{
}

/*
=================
Sys_SendKeyEvents
=================
*/
void Sys_SendKeyEvents (void)
{
	// grab frame time 
	sys_frame_time = Sys_Milliseconds();
}

#ifdef RTCTIMER_LINUX
void rtcinit() {
	int retval;
	unsigned long tmpread;

	/* try accessing rtc */
	rtc_fd = open("/dev/rtc", O_RDONLY);
	if (rtc_fd < 0) {
		Com_Printf("Could not open /dev/rtc\n");
		rtc_fd = 0;
		return;
	}

	/* make sure RTC is set to high enough rate */
	retval = ioctl(rtc_fd, RTC_IRQP_READ, &tmpread);
	if (retval < 0) {
		Com_Printf("Could not check RTC rate\n");
		close(rtc_fd);
		rtc_fd = 0;
		return;
	}
	if (tmpread < 1000) {
		Com_Printf("RTC rate is not set to high enough value\n");
		close(rtc_fd);
		rtc_fd = 0;
		return;
	}
	rtc_rate = tmpread;

	/* take ownership of rtc */
	retval = fcntl(rtc_fd, F_SETOWN, getpid());
	if (retval < 0) {
		Com_Printf("Couldn't set ownership of /dev/rtc\n");
		close(rtc_fd);
		rtc_fd = 0;
		return;
	}

	/* everything is nice - now turn on the RTC's periodic timer */
	retval = ioctl(rtc_fd, RTC_PIE_ON, 0);
	if (retval == -1) {
		Com_Printf("Error activating RTC timer\n");
		close(rtc_fd);
		rtc_fd = 0;
		return;
	}

	Com_Printf("RTC timer enabled, running at %iHz\n", rtc_rate);
}
#endif


/*****************************************************************************/

int main( int argc, char **argv )
{
	unsigned int oldtime, newtime, time;

#ifdef RTCTIMER_LINUX
	rtcinit();
#endif
	Qcommon_Init( argc, argv );

	fcntl( 0, F_SETFL, fcntl (0, F_GETFL, 0) | FNDELAY );

	nostdout = Cvar_Get( "nostdout", "0", 0 );
	if( !nostdout->integer ) {
		fcntl( 0, F_SETFL, fcntl (0, F_GETFL, 0) | FNDELAY );
	}

	oldtime = Sys_Milliseconds();
	while( qtrue )
	{
		// find time spent rendering last frame
		do {
			newtime = Sys_Milliseconds ();
			time = newtime - oldtime;
		} while( time < 1 );
		oldtime = newtime;

		Qcommon_Frame( time );
	}
}


