/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "server.h"

server_constant_t svc;				// constant server info (trully persistant since sv_init)
server_static_t	svs;				// persistant server info
server_t		sv;					// local server

//================
//SV_FindIndex
//================
int SV_FindIndex( char *name, int start, int max, qboolean create )
{
	int		i;

//	assert( name && name[0] );

	if( !name || !name[0] )
		return 0;

	if( strlen(name) >= MAX_CONFIGSTRING_CHARS )
		Com_Error( ERR_DROP, "Configstring too long: %s\n", name );

	for( i = 1; i < max && sv.configstrings[start+i][0]; i++ ) {
		if( !strncmp( sv.configstrings[start+i], name, sizeof(sv.configstrings[start+i]) ) )
			return i;
	}

	if( !create )
		return 0;

	if( i == max )
		Com_Error( ERR_DROP, "*Index: overflow" );

	Q_strncpyz( sv.configstrings[start+i], name, sizeof(sv.configstrings[i]) );

	if( sv.state != ss_loading )
	{	// send the update to everyone
		SV_SendServerCommand( NULL, "cs %i \"%s\"", start+i, name );
	}

	return i;
}


int SV_ModelIndex( char *name )
{
	return SV_FindIndex( name, CS_MODELS, MAX_MODELS, qtrue );
}

int SV_SoundIndex( char *name )
{
	return SV_FindIndex( name, CS_SOUNDS, MAX_SOUNDS, qtrue );
}

int SV_ImageIndex( char *name )
{
	return SV_FindIndex( name, CS_IMAGES, MAX_IMAGES, qtrue );
}

int SV_SkinIndex( char *name )
{
	return SV_FindIndex( name, CS_SKINFILES, MAX_SKINFILES, qtrue );
}

//================
//SV_CreateBaseline
//
//Entity baselines are used to compress the update messages
//to the clients -- only the fields that differ from the
//baseline will be transmitted
//================
void SV_CreateBaseline( void )
{
	edict_t			*svent;
	int				entnum;	

	for( entnum = 1; entnum < sv.num_edicts; entnum++ )
	{
		svent = EDICT_NUM(entnum);
#ifdef SVBASELINES_GAMECODEFILTER
		sv.baselinesServer[entnum] = svent->r;
#else // SVBASELINES_GAMECODEFILTER
		if( !svent->r.inuse )
			continue;
		if( !svent->s.modelindex && !svent->s.sound && !svent->s.effects )
			continue;
#endif // SVBASELINES_GAMECODEFILTER
		svent->s.number = entnum;

		//
		// take current state as baseline
		//
		if( !(svent->r.svflags & SVF_FORCEOLDORIGIN) )
			VectorCopy( svent->s.origin, svent->s.old_origin );
		sv.baselines[entnum] = svent->s;
	}
}

//=================
//SV_PureList_f
//=================
void SV_PureList_f( void )
{
	unsigned int i = 0;

	Com_Printf( "Pure files:\n" );
	while( i < svs.purelist_size && strlen(svs.purelist+i) )
	{
		Com_Printf( "- %s (%u)\n", svs.purelist+i, FS_PakChecksum(svs.purelist+i) );
		i += strlen(svs.purelist+i);
		i++;
	}
}

//=================
//SV_AddPurePak
//=================
static void SV_AddPurePak( const char *pakname )
{
	unsigned int i = 0;

	while( i < svs.purelist_size && strlen(svs.purelist + i) )
	{
		if( !strcmp(svs.purelist + i, pakname) )
			return;

		i += strlen( svs.purelist + i );
		i++;
	}

	while( svs.purelist_size - i < strlen(pakname) )
	{
		svs.purelist_size += 128;

		if( !svs.purelist )
			svs.purelist = Mem_Alloc( sv_mempool, svs.purelist_size );
		else
			svs.purelist = Mem_Realloc( sv_mempool, svs.purelist_size );
	}

	memcpy( svs.purelist+i, pakname, strlen(pakname) );
}

//=================
//SV_AddPureFile
//=================
static void SV_AddPureFile( const char *filename )
{
	const char *pakname;

	if( !filename || !strlen(filename) )
		return;

	pakname = FS_PakNameForFile( filename );

	if( pakname ) {
		Com_DPrintf( "Pure file: %s (%s)\n", pakname, filename );
		SV_AddPurePak( pakname );
	}
}

//=================
//SV_ReloadPureList
//=================
void SV_ReloadPureList( void )
{
	char	tempname[MAX_QPATH];
	char	buffer[20*MAX_QPATH];
	char	**paks;
	int		i, nummodules, numpaks;
	char	*s;

	if( svs.purelist )
		memset( svs.purelist, 0, svs.purelist_size );

	// *pure.pk3
	paks = NULL;
	numpaks = FS_GetExplicitPurePakList( &paks );
	if( numpaks ) {
		for ( i = 0; i < numpaks; i++ ) {
			SV_AddPurePak( paks[i] );
			Mem_ZoneFree( paks[i] );
		}
		Mem_ZoneFree( paks );
	}

	// game modules
	nummodules = FS_GetFileList( NULL, ".so", buffer, sizeof(buffer));
	s = buffer;
	for ( i = 0; i < nummodules; i++, s += strlen(s) + 1 )
		SV_AddPureFile( s );

	nummodules = FS_GetFileList( NULL, ".dll", buffer, sizeof(buffer));
	s = buffer;
	for ( i = 0; i < nummodules; i++, s += strlen(s) + 1 )
		SV_AddPureFile( s );

	// precache
	for( i = CS_MODELS; i < CS_IMAGES+MAX_IMAGES; i++ )
	{
		if( strlen(sv.configstrings[i]) == 0 )
			continue;

		if( i == CS_MODELS || i == CS_SOUNDS || i == CS_IMAGES ) // empty fields
			continue;

		if( i == CS_MODELS+1 ) // map
		{
			SV_AddPureFile( sv.configstrings[i] );
			continue;
		}
		else if( i < CS_MODELS+MAX_MODELS )
		{
			if (sv.configstrings[i][0] == '*' || sv.configstrings[i][0] == '$' || sv.configstrings[i][0] == '#')
				SV_AddPureFile( sv.configstrings[i]+1 );
			else
				SV_AddPureFile( sv.configstrings[i] );
			continue;
		}
		else if( i < CS_SOUNDS+MAX_SOUNDS )
		{
			if( sv.configstrings[i][0] == '*' ) // sexed sounds
				Q_strncpyz( tempname, sv.configstrings[i]+1, sizeof(tempname) );
			else
				Q_strncpyz( tempname, sv.configstrings[i], sizeof(tempname) );
			COM_DefaultExtension( tempname, ".wav", sizeof(tempname) );
			SV_AddPureFile( tempname );
			continue;
		}
		else if( i < CS_IMAGES+MAX_IMAGES )
		{
			// server doesn't know about shaders...
			continue;
		}
	}
}

//================
//SV_SpawnServer
//Change the server to a new map, taking all connected clients along with it.
//================
void SV_SpawnServer( const char *server, qboolean devmap )
{
	unsigned	checksum;

	if( devmap )
		Cvar_ForceSet( "sv_cheats", "1" );
	Cvar_FixCheatVars();

	Com_Printf( "------- Server Initialization -------\n" );
	Com_DPrintf( "SpawnServer: %s\n",server );

	svs.spawncount++;		// any partially connected client will be restarted

	Com_SetServerState( ss_dead );

	// wipe the entire per-level structure
	memset( &sv, 0, sizeof(sv) );
	SV_ResetClientFrameCounters();
	svs.realtime = 0;
	svs.gametime = 0;

	Q_strncpyz( sv.name, server, sizeof(sv.name) );

	sv.nextSnapTime = 1000;
	
	// load the map
	Q_snprintfz( sv.configstrings[CS_MODELS+1], sizeof(sv.configstrings[CS_MODELS+1]), "maps/%s.bsp", server );
	CM_LoadMap( sv.configstrings[CS_MODELS+1], qfalse, &checksum );

	Q_snprintfz( sv.configstrings[CS_MAPCHECKSUM], sizeof(sv.configstrings[CS_MAPCHECKSUM]), "%i", checksum );

	//
	// spawn the rest of the entities on the map
	//	

	// precache and static commands can be issued during
	// map initialization
	sv.state = ss_loading;
	Com_SetServerState( sv.state );

	// load and spawn all other entities
	ge->SpawnEntities( sv.name, CM_EntityString(), CM_EntityStringLen() );

	// run two frames to allow everything to settle
	ge->RunFrame( svc.snapFrameTime, svs.gametime );
	ge->RunFrame( svc.snapFrameTime, svs.gametime );
#ifdef SVBASELINES_GAMECODEFILTER
	ge->SnapFrame(); // set up the first snap so baselines start from a valid point
#endif
	SV_CreateBaseline(); // create a baseline for more efficient communications

	// all precaches are complete
	sv.state = ss_game;
	Com_SetServerState( sv.state );

	// set serverinfo variable
	Cvar_FullSet( "mapname", sv.name, CVAR_SERVERINFO | CVAR_READONLY, qtrue );

	// set purelist
	SV_ReloadPureList();

	Com_Printf( "-------------------------------------\n" );
}

//==============
//SV_InitGame
//A brand new game has been started
//==============
void SV_InitGame( void )
{
	int		i;
	edict_t	*ent;

	if( svs.initialized )
	{
		// cause any connected clients to reconnect
		SV_Shutdown( "Server restarted\n", qtrue );

		// SV_Shutdown will also call Cvar_GetLatchedVars
	}
	else
	{
		// make sure the client is down
		CL_Disconnect( NULL );
		SCR_BeginLoadingPlaque();

		// get any latched variable changes (sv_maxclients, etc)
		Cvar_GetLatchedVars( CVAR_LATCH );
	}

	svs.initialized = qtrue;

	// init clients
	if( sv_maxclients->integer <= 1 )
		Cvar_FullSet( "sv_maxclients", "8", CVAR_SERVERINFO | CVAR_LATCH, qtrue );
	else if( sv_maxclients->integer > MAX_CLIENTS )
		Cvar_FullSet( "sv_maxclients", va("%i", MAX_CLIENTS), CVAR_SERVERINFO | CVAR_LATCH, qtrue );

	svs.spawncount = rand();
	svs.clients = Mem_Alloc( sv_mempool, sizeof(client_t)*sv_maxclients->integer );
	svs.num_client_entities = sv_maxclients->integer * UPDATE_BACKUP * MAX_SNAP_ENTITIES;
	svs.client_entities = Mem_Alloc( sv_mempool, sizeof(entity_state_t) * svs.num_client_entities );

	// init network stuff
	NET_Config( (sv_maxclients->integer > 1) );

	// init game
	SV_InitGameProgs();
	for( i=0; i < sv_maxclients->integer; i++ )
	{
		ent = EDICT_NUM(i+1);
		ent->s.number = i+1;
		svs.clients[i].edict = ent;
		memset( &svs.clients[i].lastcmd, 0, sizeof(svs.clients[i].lastcmd) );
	}
}

//======================
//SV_Map
// command from the console or progs.
//======================
void SV_Map( const char *level, qboolean devmap )
{
	client_t	*cl;
	int			i;

	if( svs.demofile )
		SV_DemoRecordStop_f();

	// skip the end-of-unit flag if necessary
	if( level[0] == '*' )
		level++;

	if( sv.state == ss_dead )
		SV_InitGame();	// the game is just starting

	// remove all bots before changing map
	for( i = 0, cl = svs.clients; i < sv_maxclients->integer; i++, cl++ ) {
		if( cl->state && cl->edict && (cl->edict->r.svflags & SVF_FAKECLIENT) ) {
			SV_DropClient( cl, DROP_TYPE_GENERAL, "SV_Map" );
		}
	}

	// wsw : Medar : this used to be at SV_SpawnServer, but we need to do it before sending changing
	//               so we don't send frames after sending changing command
	// leave slots at start for clients only
	for( i = 0; i < sv_maxclients->integer; i++ )
	{
		// needs to reconnect
		if( svs.clients[i].state > CS_CONNECTED )
			svs.clients[i].state = CS_CONNECTED;
		svs.clients[i].lastframe = -1;
		memset( svs.clients[i].gameCommands, 0, sizeof(svs.clients[i].gameCommands) );
	}

	SCR_BeginLoadingPlaque();			// for local system
	SV_BroadcastCommand( "changing\n" );
	SV_SendClientMessages();
	SV_SpawnServer( level, devmap );
#ifndef DISABLE_DEFER
	Cbuf_CopyToDefer();
#endif

	SV_BroadcastCommand( "reconnect\n" );
}
