/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "server.h"

mempool_t	*sv_mempool;

cvar_t	*sv_enforcetime;

cvar_t	*sv_timeout;			// seconds without any message
cvar_t	*sv_zombietime;			// seconds to sink messages after disconnect

cvar_t	*rcon_password;			// password for remote server commands

cvar_t	*sv_uploads;
cvar_t	*sv_uploads_from_server;
cvar_t	*sv_uploads_baseurl;

cvar_t	*sv_pure;

cvar_t	*sv_noreload;			// don't reload level state when reentering

cvar_t	*sv_maxclients;
cvar_t	*sv_showclamp;

cvar_t	*sv_hostname;
cvar_t	*sv_public;			// should heartbeats be sent

cvar_t	*sv_reconnectlimit;	// minimum seconds between connect messages

// wsw : jal

cvar_t	*sv_maxrate;
cvar_t	*sv_compresspackets;
cvar_t	*sv_masterservers;
cvar_t	*sv_skilllevel;

// wsw : debug netcode
cvar_t	*sv_debug_serverCmd;

#ifdef BATTLEYE
cvar_t	*sv_battleye;
#endif

//============================================================================

//===================
//SV_CalcPings
//
//Updates the cl->ping variables
//===================
void SV_CalcPings( void )
{
	int			i;
	client_t	*cl;
	int			total, count, j;

	for( i = 0 ; i < sv_maxclients->integer; i++ )
	{
		cl = &svs.clients[i];
		if( cl->state != CS_SPAWNED )
			continue;
		if( cl->edict && (cl->edict->r.svflags & SVF_FAKECLIENT) )
			continue;
		total = 0;
		count = 0;
		for( j = 0; j<LATENCY_COUNTS; j++ )
		{
			if( cl->frame_latency[j] > 0 )
			{
				count++;
				total += cl->frame_latency[j];
			}
		}
		if( !count )
			cl->ping = 0;
		else
#if 0
			cl->ping = total*100/count - 100;
#else
			cl->ping = total / count;
#endif
		// let the game dll know about the ping
		cl->edict->r.client->r.ping = cl->ping;
	}
}

//===================
//SV_GiveMsec
//
//Every few frames, gives all clients an allotment of milliseconds
//for their command moves.  If they exceed it, assume cheating.
//===================
void SV_GiveMsec( void )
{
	int			i;
	client_t	*cl;

	if( sv.framenum & 15 )
		return;

	for( i = 0 ; i < sv_maxclients->integer; i++ )
	{
		cl = &svs.clients[i];
		if( cl->state == CS_FREE )
			continue;
		
		cl->commandMsec = 1800;		// 1600 + some slop
	}
}

//=================
// SV_ProcessPacket
//=================
qboolean SV_ProcessPacket( netchan_t *netchan, msg_t *msg ) {
	int			sequence, sequence_ack;
	int			qport = -1;
	int			zerror;

	if( !Netchan_Process( netchan, msg ) )
		return qfalse;		// wasn't accepted for some reason

	// now if compressed, expand it
	MSG_BeginReading( msg );
	sequence = MSG_ReadLong( msg );
	sequence_ack = MSG_ReadLong( msg ); 
	qport = MSG_ReadShort( msg );
	if( msg->compressed ) {
		zerror = Netchan_DecompressMessage( msg );
		if( zerror < 0 ) { // compression error. Drop the packet
			Com_DPrintf( "SV_ProcessPacket: Compression error %i. Dropping packet\n", zerror );
			return qfalse;
		}
	}

	return qtrue;
}

//=================
//SV_ReadPackets
//=================
void SV_ReadPackets( void )
{
	int			i;
	client_t	*cl;
	int			qport;

	static msg_t		msg;
	static qbyte		msgData[MAX_MSGLEN];

	MSG_Init( &msg, msgData, sizeof(msgData) );
	MSG_Clear( &msg );

	while( NET_GetPacket( NS_SERVER, &net_from, &msg ) )
	{
		// check for connectionless packet (0xffffffff) first
		if( *(int *)msg.data == -1 ) {
			SV_ConnectionlessPacket( &msg );
			continue;
		}

		// read the qport out of the message so we can fix up
		// stupid address translating routers
		MSG_BeginReading( &msg );
		MSG_ReadLong( &msg );		// sequence number
		MSG_ReadLong( &msg );		// sequence number
		qport = MSG_ReadShort( &msg ) & 0xffff;
		// data follows

		// check for packets from connected clients
		for( i = 0, cl = svs.clients; i < sv_maxclients->integer; i++, cl++ )
		{
			if( cl->state == CS_FREE )
				continue;
			if( cl->edict && (cl->edict->r.svflags & SVF_FAKECLIENT) )
				continue;
			if( !NET_CompareBaseAdr(&net_from, &cl->netchan.remoteAddress) )
				continue;
			if( cl->netchan.qport != qport )
				continue;
			if( cl->netchan.remoteAddress.port != net_from.port )
			{
				Com_Printf( "SV_ReadPackets: fixing up a translated port\n" );
				cl->netchan.remoteAddress.port = net_from.port;
			}

			if( SV_ProcessPacket( &cl->netchan, &msg ) )
			{	// this is a valid, sequenced packet, so process it
				if( cl->state != CS_ZOMBIE )
				{
					cl->lastPacketReceivedTime = svs.realtime;	// don't timeout
					SV_ParseClientMessage( cl, &msg );
				}
			}
			break;
		}
		
		if( i != sv_maxclients->integer )
			continue;
	}
}

//==================
//SV_CheckTimeouts
//
//If a packet has not been received from a client for timeout->value
//seconds, drop the conneciton.  Server frames are used instead of
//realtime to avoid dropping the local client while debugging.
//
//When a client is normally dropped, the client_t goes into a zombie state
//for a few seconds to make sure any final reliable message gets resent
//if necessary
//==================
void SV_CheckTimeouts( void )
{
	int		i;
	client_t	*cl;

	for( i = 0, cl = svs.clients; i < sv_maxclients->integer; i++, cl++ )
	{
		// fake clients do not timeout
		if( cl->edict && (cl->edict->r.svflags & SVF_FAKECLIENT) )
			cl->lastPacketReceivedTime = svs.realtime;

		// message times may be wrong across a changelevel
		else if( cl->lastPacketReceivedTime > svs.realtime )
			cl->lastPacketReceivedTime = svs.realtime;

		if( cl->state == CS_ZOMBIE && cl->lastPacketReceivedTime + 1000 * sv_zombietime->value < svs.realtime )
		{
			cl->state = CS_FREE;	// can now be reused
			continue;
		}

		if( (cl->state != CS_FREE && cl->state != CS_ZOMBIE) &&
			(cl->lastPacketReceivedTime + 1000 * sv_timeout->value < svs.realtime ) )
		{
			SV_ClientPrintf( NULL, "%s%s timed out\n", cl->name, S_COLOR_WHITE );
			SV_DropClient( cl, DROP_TYPE_GENERAL, "Connection timed out" );
			cl->state = CS_FREE;	// don't bother with zombie state
		}

		// timeout downloads left open
		if( (cl->state != CS_FREE && cl->state != CS_ZOMBIE) &&
			(cl->download.name && cl->download.timeout < svs.realtime) )
		{
			Com_Printf( "Download of %s to %s%s timed out\n", cl->download.name, cl->name, S_COLOR_WHITE );

			if( cl->download.data ) {
				FS_FreeBaseFile( cl->download.data );
				cl->download.data = NULL;
			}

			Mem_ZoneFree( cl->download.name );
			cl->download.name = NULL;

			cl->download.size = 0;
			cl->download.timeout = 0;
		}
	}
}

//#define WORLDFRAMETIME 25 // 40fps
//#define WORLDFRAMETIME 20 // 50fps
#define WORLDFRAMETIME 16 // 62.5fps
//=================
//SV_RunGameFrame
//=================
qboolean SV_RunGameFrame( int msec )
{
	int extraTime = 0;
	static int accTime = 0;

	accTime += msec;

	// move autonomous things around if enough time has passed
	if( svs.gametime < sv.nextSnapTime )
	{
		if( svs.gametime + svc.snapFrameTime < sv.nextSnapTime ) {
			if( sv_showclamp->integer )
				Com_Printf( "sv lowclamp\n" );
			sv.nextSnapTime = svs.gametime + svc.snapFrameTime;
			return qfalse;
		}

		// see if it's time to advance the world
		if( accTime >= WORLDFRAMETIME ) {
			ge->RunFrame( accTime, svs.gametime );
			accTime = 0;
		}

		if( !SV_SendClientsFragments() )
			NET_Sleep( sv.nextSnapTime - svs.gametime ); // FIXME: gametime might slower/faster than real time

		return qfalse;
	}

	// update ping based on the last known frame from all clients
	SV_CalcPings();

	if( sv.nextSnapTime < svs.gametime ) {
		extraTime = (int)(svs.gametime - sv.nextSnapTime);
	}
	if( extraTime >= msec )
		extraTime = msec - 1;

	sv.nextSnapTime = svs.gametime + svc.snapFrameTime - extraTime;

	if( host_speeds->integer )
		time_before_game = Sys_Milliseconds();

	// Execute all clients pending move commands
	if( accTime ) {
		ge->RunFrame( accTime, svs.gametime );
		accTime = 0;
	}

	sv.framenum++;
	ge->SnapFrame();

	if( host_speeds->integer )
		time_after_game = Sys_Milliseconds();

	return qtrue;
}

//==================
//SV_Frame
//==================
void SV_Frame( int realmsec, int gamemsec )
{
	time_before_game = time_after_game = 0;

	// if server is not active, do nothing
	if( !svs.initialized )
		return;

	svs.realtime += realmsec;
	svs.gametime += gamemsec;

	// check timeouts
	SV_CheckTimeouts();

	// get packets from clients
	SV_ReadPackets();

	// let everything in the world think and move
	if( SV_RunGameFrame(gamemsec) ) {

		// give the clients some timeslices
		SV_GiveMsec();

		// send messages back to the clients that had packets read this frame
		SV_SendClientMessages();

		// save the entire world state if recording a serverdemo
		SV_DemoRecordMessage();

		// send a heartbeat to the master if needed
		SV_MasterHeartbeat();

		// clear teleport flags, etc for next frame
		ge->ClearSnap();
	}

#ifdef BATTLEYE
	// check for connection to and incoming packets from BattlEye Master Server
	if (sv_battleye->integer)
		SV_BattlEyeFrame(realmsec);
#endif
}

#ifdef BATTLEYE
//==================
//SV_BattlEyeFrame
//==================
void SV_BattlEyeFrame( int msec )
{
	// connectstate: -1 = disconnected, 0 = connecting, 1 = connected
	static int connectstate = -1;
	static int nextconnect_countdown = 0;
	int i;
	client_t* cl;

	if (connectstate == -1)
	{
		if ((nextconnect_countdown -= msec) <= 0)
		{
			// only use BE if there are BE-enabled players on the server
			for (i = 0; i < sv_maxclients->integer; i++)
			{
				cl = &svs.clients[i];
				if (cl->state == CS_SPAWNED && cl->battleye)
				{
					connectstate = (SV_BE_ConnectToMaster() ? 0 : -1);
					break;
				}
			}
		}
		// do not re-set nextconnect_countdown below
		else
			return;
	}
	else if (connectstate == 0)
	{
		connectstate = SV_BE_CheckConnectAttempt();

		if (connectstate == 1)
		{
			char version_packet[20];

			Com_Printf("Connected to BattlEye Master Server\n");

			// send game version info
			version_packet[0] = (char)sprintf(version_packet+1, "%.3f", VERSION) + 1;
			SV_BE_SendToMaster(version_packet, version_packet[0]+1);
		}
		else if (connectstate == -1)
			Com_Printf("Couldn't connect to BattlEye Master Server: %s\n", NET_ErrorString());
	}
	else
	{
		qbyte packetid;
		int result;
		qboolean firstreceived = qfalse;

		for (i = 0; i < sv_maxclients->integer; i++)
		{
			cl = &svs.clients[i];
			if (cl->battleye && ((!cl->addedToBEMaster && cl->state == CS_SPAWNED)
				|| (cl->addedToBEMaster && cl->state < CS_SPAWNED)))
			{
				// inform BE Master about new/removed player
				qbyte changeplayer_packet[2];
				changeplayer_packet[0] = 1; /*packet type*/
				changeplayer_packet[1] = cl - svs.clients; /*player id*/
				SV_BE_SendToMaster(changeplayer_packet, sizeof(changeplayer_packet));
				cl->addedToBEMaster = !cl->addedToBEMaster;
			}
		}

		// result: -1 = failed or not enough data (!= param len), 0 = no data, 1 = succeeded
		if ((result = SV_BE_ReceiveFromMaster(&packetid, sizeof(packetid))) == 1)
		{
			firstreceived = qtrue;

			// distribute packet to all connected (human) players
			if (packetid == 0)
			{
				short packetlen;
				static qbyte packet[BE_MAX_PACKET_SIZE];

				// receive actual packet
				if ((result = SV_BE_ReceiveFromMaster(&packetlen, sizeof(packetlen))) != 1) goto disconnect;
				if (packetlen <= 0 || packetlen > BE_MAX_PACKET_SIZE) { result = -1; goto disconnect; }
				if ((result = SV_BE_ReceiveFromMaster(packet, packetlen)) != 1) goto disconnect;

				for (i = 0; i < sv_maxclients->integer; i++)
				{
					cl = &svs.clients[i];
					if (cl->addedToBEMaster)
					{
						// send packet to client
						memcpy(cl->BE.packets[cl->BE.headPacket & BE_UPDATE_MASK], packet, packetlen);
						cl->BE.packetLens[cl->BE.headPacket & BE_UPDATE_MASK] = packetlen;
						cl->BE.headPacket++;
					}
				}
			}
			// violation: display a server message and kick specified player (with that msg)
			else if (packetid == 1)
			{
				qbyte playerid;
				qbyte messagelen;
				static char message[255];

				if ((result = SV_BE_ReceiveFromMaster(&playerid, sizeof(playerid))) != 1) goto disconnect;
				if (playerid >= sv_maxclients->integer) { result = -1; goto disconnect; }
				if ((result = SV_BE_ReceiveFromMaster(&messagelen, sizeof(messagelen))) != 1) goto disconnect;
				if ((result = SV_BE_ReceiveFromMaster(message, messagelen)) != 1) goto disconnect;

				cl = &svs.clients[playerid];
				if (cl->addedToBEMaster)
				{
					SV_ClientPrintf(NULL, "%s" S_COLOR_WHITE " was kicked by BattlEye: " S_COLOR_YELLOW "%s\n", cl->name, message);
					SV_DropClient(cl, DROP_TYPE_NORECONNECT, "BattlEye: %s", message);
				}
			}
		}
disconnect:
		// disconnect if failed, not enough or no at all (after first recv call) returned data
		if (result == -1 || (firstreceived && result == 0))
		{
			SV_BE_DisconnectFromMaster();
			Com_Printf("Disconnected from BattlEye Master Server\n");
			connectstate = -1;
			for (i = 0; i < sv_maxclients->integer; i++)
				svs.clients[i].addedToBEMaster = qfalse;
		}
	}
	// when disconnected, try to reconnect every 60 secs
	if (connectstate == -1)
		nextconnect_countdown = 60000;
}
#endif

//============================================================================


//=================
//SV_UserinfoChanged
//
//Pull specific info from a newly changed userinfo string
//into a more C friendly form.
//=================
void SV_UserinfoChanged( client_t *cl )
{
	char	*val;
	char	*s;
	int		i;

	// wsw : r1q2 : check userinfo string for end-of-message-in-string exploit
	if( strchr (cl->userinfo, '\xFF') )
	{
		SV_DropClient( cl, DROP_TYPE_GENERAL, "Tried to use end-of-message-in-string exploit" );
		return;
	}
	
	// wsw : jal : validate model and skin names ( can't contain "/" )
	// skin
	val = Info_ValueForKey( cl->userinfo, "skin" );
	if( strchr(val, '/') ) {
		Com_Printf( "SV_UserinfoChanged: Fixing invalid skin name %s\n", val );
		s = strchr(val, '/');
		*s = 0;
	}
	if( !strlen(val) )
		val = DEFAULT_PLAYERSKIN;
	
	Info_SetValueForKey( cl->userinfo, "skin", val );
	
	// model
	val = Info_ValueForKey( cl->userinfo, "model" );
	if( strchr(val, '/') ) {
		Com_Printf( "SV_UserinfoChanged: Fixing invalid model name %s\n", val );
		s = strchr(val, '/');
		*s = 0;
	}
	if( !strlen(val) )
		val = DEFAULT_PLAYERMODEL;
	
	Info_SetValueForKey( cl->userinfo, "model", val );

	// call prog code to allow overrides
	ge->ClientUserinfoChanged( cl->edict, cl->userinfo );

	// wsw : r1q2[start] : verify, validate, truncate and print name changes
	val = Info_ValueForKey( cl->userinfo, "name" );
	if( !val || !val[0] ) {
		Q_strncpyz( cl->name, "Player", sizeof(cl->name) );
	} else {
		Q_strncpyz( cl->name, val, sizeof(cl->name) );

		// mask off high bit
		for( i = 0; i < sizeof(cl->name); i++ )
			cl->name[i] &= 127;
	}

	// rate command
	if( Sys_IsLANAddress( cl->netchan.remoteAddress ) && sv_public->integer != 1 ) { 
#ifndef RATEKILLED
		cl->rate = 99999;	// lans should not rate limit
#endif
	} else {
		val = Info_ValueForKey( cl->userinfo, "rate" );
		if( strlen(val) )
		{
			int newrate;

			newrate = atoi(val);
			if( sv_maxrate->integer && newrate > sv_maxrate->integer )
				newrate = sv_maxrate->integer;
			else if( newrate > 90000 )
				newrate = 90000;
			if( newrate < 1000 )
				newrate = 1000;
#ifndef RATEKILLED
			if( cl->rate != newrate ) {
				cl->rate = newrate;
				Com_Printf( "%s%s has rate %i\n", cl->name, S_COLOR_WHITE, cl->rate );
			}
#endif
		}
#ifndef RATEKILLED
		else
			cl->rate = 5000;
#endif
	}
}


//============================================================================

//===============
//SV_Init
//
//Only called at plat.exe startup, not for each game
//===============
void SV_Init( void )
{
	cvar_t		*sv_pps;
	cvar_t		*sv_fps;

	SV_InitOperatorCommands();

	sv_mempool = Mem_AllocPool( NULL, "Server" );

	Cvar_Get( "dmflags", 0, CVAR_SERVERINFO );
	Cvar_Get( "sv_cheats", "0", CVAR_SERVERINFO | CVAR_LATCH );
	Cvar_Get( "protocol", va("%i", PROTOCOL_VERSION), CVAR_SERVERINFO | CVAR_NOSET );

	rcon_password =				Cvar_Get( "rcon_password", "", 0 );
	sv_hostname =				Cvar_Get( "sv_hostname", "warsow server", CVAR_SERVERINFO | CVAR_ARCHIVE );
	sv_timeout =				Cvar_Get( "sv_timeout", "125", 0 );
	sv_zombietime =				Cvar_Get( "sv_zombietime", "2", 0 );
	sv_showclamp =				Cvar_Get( "sv_showclamp", "0", 0 );
	sv_enforcetime =			Cvar_Get( "sv_enforcetime", "0", 0 );

	sv_uploads =				Cvar_Get( "sv_uploads", "1", CVAR_ARCHIVE );
	sv_uploads_from_server =	Cvar_Get( "sv_uploads_from_server", "1", CVAR_ARCHIVE );
	sv_uploads_baseurl =		Cvar_Get( "sv_uploads_baseurl", "", CVAR_ARCHIVE );

	if( dedicated->integer )
		sv_pure =					Cvar_Get( "sv_pure", "1", CVAR_ARCHIVE | CVAR_LATCH | CVAR_SERVERINFO );
	else
		sv_pure =					Cvar_Get( "sv_pure", "0", CVAR_ARCHIVE | CVAR_LATCH | CVAR_SERVERINFO );

#ifdef BATTLEYE
	if( dedicated->integer )
		sv_battleye =			Cvar_Get( "sv_battleye", "1", CVAR_SERVERINFO | CVAR_ARCHIVE | CVAR_LATCH );
	else
		sv_battleye =			Cvar_Get( "sv_battleye", "0", CVAR_SERVERINFO | CVAR_READONLY );
#endif

	sv_noreload =				Cvar_Get( "sv_noreload", "0", 0 );
	
	if( dedicated->integer )
		sv_public =				Cvar_Get( "sv_public", "1", CVAR_ARCHIVE|CVAR_NOSET );
	else
		sv_public =				Cvar_Get( "sv_public", "0", CVAR_ARCHIVE );
	sv_reconnectlimit =			Cvar_Get( "sv_reconnectlimit", "3", CVAR_ARCHIVE );
	sv_maxclients =				Cvar_Get( "sv_maxclients", "1", CVAR_ARCHIVE | CVAR_SERVERINFO | CVAR_LATCH );

	// fix invalid sv_maxclients values
	if( sv_maxclients->integer < 1 )
		Cvar_FullSet( "sv_maxclients", "1", CVAR_ARCHIVE|CVAR_SERVERINFO|CVAR_LATCH, qtrue );
	else if( sv_maxclients->integer > MAX_CLIENTS )
		Cvar_FullSet( "sv_maxclients", va("%i", MAX_CLIENTS), CVAR_ARCHIVE|CVAR_SERVERINFO|CVAR_LATCH, qtrue );

	// wsw : jal : cap client's exceding server rules
	sv_maxrate =				Cvar_Get( "sv_maxrate", "15000", CVAR_SERVERINFO|CVAR_ARCHIVE );
	sv_compresspackets =		Cvar_Get( "sv_compresspackets", "1", CVAR_DEVELOPER );
	sv_skilllevel =				Cvar_Get( "sv_skilllevel", "1", CVAR_SERVERINFO|CVAR_ARCHIVE );
	if( dedicated->integer )
		sv_masterservers =			Cvar_Get( "masterservers", DEFAULT_MASTER_SERVERS_IPS, CVAR_NOSET );
	else
		sv_masterservers =			Cvar_Get( "masterservers", DEFAULT_MASTER_SERVERS_IPS, 0 );
	sv_debug_serverCmd =		Cvar_Get( "sv_debug_serverCmd", "0", CVAR_ARCHIVE );

	// this is a message holder for shared use
	MSG_Init( &tmpMessage, tmpMessageData, sizeof(tmpMessageData) );

	// init server updates ratio
	if( dedicated->integer )
		sv_pps = Cvar_Get( "sv_pps", "20", CVAR_SERVERINFO|CVAR_NOSET );
	else
		sv_pps = Cvar_Get( "sv_pps", "20", CVAR_SERVERINFO );
	svc.snapFrameTime = (int)( 1000 / sv_pps->value );
	if( svc.snapFrameTime > 200 ) {	// too slow, also, netcode uses a byte
		Cvar_ForceSet( "sv_pps", "5" );
		svc.snapFrameTime = 200;
	} else if( svc.snapFrameTime < 10 ) { // abusive
		Cvar_ForceSet( "sv_pps", "100" );
		svc.snapFrameTime = 10;
	}

	sv_fps = Cvar_Get( "sv_fps", "62", CVAR_NOSET );
	svc.gameFrameTime = (int)( 1000 / sv_fps->value );
	if( svc.gameFrameTime > svc.snapFrameTime ) { // gamecode can never be slower than snaps
		svc.gameFrameTime = svc.snapFrameTime;
		Cvar_ForceSet( "sv_fps", sv_pps->dvalue );
	}

	Com_Printf( "Game running at %i fps. Server transmit at %i pps\n", sv_fps->integer, sv_pps->integer );

	//init the master servers list
	SV_InitMaster();
}

//==================
//SV_FinalMessage
//
//Used by SV_Shutdown to send a final message to all
//connected clients before the server goes down.  The messages are sent immediately,
//not just stuck on the outgoing message list, because the server is going
//to totally exit after returning from this function.
//==================
void SV_FinalMessage( char *message, qboolean reconnect )
{
	int			i, j;
	client_t	*cl;

	for( i = 0, cl = svs.clients; i < sv_maxclients->integer; i++, cl++ ) {
		if( cl->edict && (cl->edict->r.svflags & SVF_FAKECLIENT) )
			continue;
		if( cl->state >= CS_AWAITING )
		{
			if( reconnect )
				SV_SendServerCommand( cl, "forcereconnect \"%s\"", message );
			else
				SV_SendServerCommand( cl, "disconnect %i \"%s\"", DROP_TYPE_GENERAL, message );

			MSG_Clear( &tmpMessage );
			SV_AddReliableCommandsToMessage( cl, &tmpMessage );
			for( j = 0; j < 2; j++ ) { // send it twice
				SV_SendMessageToClient( cl, &tmpMessage );
			}
		}
	}
}

//================
//SV_Shutdown
//
//Called when each game quits,
//before Sys_Quit or Sys_Error
//================
void SV_Shutdown( char *finalmsg, qboolean reconnect )
{
	if( svs.demofile )
		SV_DemoRecordStop_f();

	if( svs.clients )
		SV_FinalMessage( finalmsg, reconnect );

	SV_ShutdownGameProgs();

	// get any latched variable changes (sv_maxclients, etc)
	Cvar_GetLatchedVars( CVAR_LATCH );

	memset( &sv, 0, sizeof(sv) );
	Com_SetServerState( sv.state );

	if( svs.purelist ) {
		Mem_Free( svs.purelist );
		svs.purelist = NULL;
		svs.purelist_size = 0;
	}

	if( sv_mempool )
		Mem_EmptyPool( sv_mempool );
	memset( &svs, 0, sizeof(svs) );

#ifdef BATTLEYE
	if( sv_battleye->integer )
		SV_BE_DisconnectFromMaster();
#endif
}
